/***************************************************************************
 *   Copyright (C) 2006 by Michael Schulze                                 *
 *   mike.s@genion.de                                                      *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/
#include "consistencycheck.h"

#include "fileutils.h"

using namespace itunesdb;

const QString ConsistencyCheck::defaultmimetype = QString( "text/html");
const QString ConsistencyCheck::name = QString( "Consistency Check");

ConsistencyCheck::ConsistencyCheck()
 : IPodUtility()
{
}


ConsistencyCheck::~ConsistencyCheck()
{
}


const QString& ConsistencyCheck::getName() {
    return name;
}


const QString& ConsistencyCheck::getDefaultMimeType() {
    return defaultmimetype;
}


class MissingFileTrackPredicate : public TrackPredicate {
    const ITunesDB& m_itunesdb;
public:
    MissingFileTrackPredicate( const ITunesDB& itunesdb ) :
        m_itunesdb( itunesdb ) {}

    virtual ~MissingFileTrackPredicate() {}

    bool operator () ( const Track * track ) const {
        return track && !QFile( m_itunesdb.getFileForPathInfo( track->getPathInfo() ) ).exists();
    }

};

void ConsistencyCheck::printMissingFilesTable( const ITunesDB& itunesdb, QTextStream& stream ) {
    // get the tracks that reference to non existing files
    ITunesDB::FilteredTrackConstIterator<MissingFileTrackPredicate> missingTracksIter =
            itunesdb.getTracksBy( MissingFileTrackPredicate( itunesdb ) );

    stream << "Missing Files" << endl;
    if ( missingTracksIter.hasNext() ) {
        stream << "<table border=\"1\"><th>Artist - Album - Title</th><th>missing File</th>" << endl;
        uint numFiles = 0;
        while ( missingTracksIter.hasNext() ) {
            ITunesDBTrack * track = missingTracksIter.next();
            stream << "<tr><td><pre>";
            stream << track->getArtist() << "\n" << track->getAlbum() << "\n" <<  track->getTitle() << "</pre></td>";
            stream << "<td><pre>" << track->getFilePath() << "</pre></td></tr>" << endl;
            numFiles++;
        }
        stream << "</table>" << endl;
        stream << "total: " << numFiles << "<br/>" << endl;
    }
}

void ConsistencyCheck::removeMissingFileTracks( KPod& ipod ) {
    // get the tracks that reference to non existing files
    ITunesDB& itunesdb = ipod.getITunesDB();
    ITunesDB::FilteredTrackConstIterator<MissingFileTrackPredicate> missingTracksIter =
            itunesdb.getTracksBy( MissingFileTrackPredicate( itunesdb ) );

    while ( missingTracksIter.hasNext() ) {
        ipod.deleteTrack( missingTracksIter.next()->getID() );
    }
}

void ConsistencyCheck::printMultipleReferencedFiles( const ITunesDB& itunesdb, QTextStream& stream ) {

    MultipleTrackDict multipleReferenced( itunesdb.getNumTracks(), false );
    multipleReferenced.setAutoDelete( true );
    ITunesDB::TrackConstIterator allTracksIter = itunesdb.getAllTracks();
    while ( allTracksIter.hasNext() ) {
        ITunesDBTrack * track = allTracksIter.next();
        TrackPtrList * sameFileList = multipleReferenced.find( track->getFilePath() );
        if ( !sameFileList ) {
            sameFileList = new TrackPtrList();
            multipleReferenced.insert( track->getFilePath(), sameFileList );
        }
        sameFileList->append( track );
    }

    MultipleTrackDictIter fileByTrackIter( multipleReferenced );

    stream << "Multiple referenced files" << endl;
    stream << "<table border=\"1\"><th>referenced File</th><th>Artist - Album - Title</th>" << endl;
    for ( ; fileByTrackIter.current(); ++fileByTrackIter ) {
        TrackPtrList * sameFileList = fileByTrackIter.current();
        QString filePath = fileByTrackIter.currentKey();
        if ( QFile( filePath ).exists() && sameFileList->count() > 1 ) {
            stream << "<tr><td>" << filePath << "</td><td>";
            TrackPtrList::Iterator trackiter = sameFileList->iterator();
            while ( trackiter.hasNext() ) {
                Track * track = trackiter.next();
                stream << track->getArtist() << " - " << track->getAlbum() << " - " << track->getTitle() << "<br/>";
            }
            stream << "</td></tr>" << endl;
        }
    }
    stream << "</table><br/>" << endl;

}


void ConsistencyCheck::checkForUnreferencedFiles( KPod& ipod, QTextStream& stream ) {
    QStringList unreferenced;
    ipod.findUnreferencedMusicFiles( unreferenced );

    stream << "<p/>Unreferenced Files<br/>" << endl;

    ITunesDB& itunesdb = ipod.getITunesDB();
    QStringList::iterator unrefIter = unreferenced.begin();
    for( ; unrefIter != unreferenced.end(); ++unrefIter ) {
        ITunesDBTrack * track = itunesdb.createNewTrack();
        track->setPathInfo( *unrefIter );
        QString filePath = track->getFilePath();

        if ( !FileUtils::readMP3Metadata( track->getFilePath(), track ) ) {
#ifdef CONSISTENCY_CHECK_ENSURE_CONSISTENCY
            QFile( filePath ).remove();
            stream << "could not read meta info from file " << filePath << ": deleted.<br/>";
#else
            stream << "could not read meta info from file " << filePath << ".<br/>";
#endif
            continue;
        }

        if ( itunesdb.findTrack( track->getArtist(), track->getAlbum(), track->getTitle(), track->getTrackNumber() ) ) {
#ifdef CONSISTENCY_CHECK_ENSURE_CONSISTENCY
            QFile( filePath ).remove();
            stream << track->toQString() << " already exists: deleted.<br/>";
#else
            stream << track->toQString() << " already exists<br/>";
#endif
            continue;
        }

        if (  ipod.addTrack( track ) != KPod::Err_None ) {
#ifdef CONSISTENCY_CHECK_ENSURE_CONSISTENCY
            QFile( filePath ).remove();
            stream << "could not add file " << filePath << ": deleted.<br/>";
#else
            stream << "could not add file " << filePath << ".<br/>";
#endif
            continue;
        }

        stream << filePath << " successfully added as " << track->getArtist() << " - " << track->getAlbum() << " - " << track->getTitle() << "<br/>" << endl;

    }
}


void ConsistencyCheck::handleRequest( KPod& ipod, const KURL& url, QByteArray& response, QString& mimeType ) {

    ITunesDB& itunesdb = ipod.getITunesDB();

    QTextOStream stream( response );
    QString title= ipod.getName();

    mimeType= QString( "text/html" );
    stream << "<html><head><title>" + name + "</title></head>\n<body bgcolor=\"white\">";

    // multiple referenced files table
    printMultipleReferencedFiles( itunesdb, stream );

    // missing files table
    printMissingFilesTable( itunesdb, stream );

#ifdef CONSISTENCY_CHECK_ENSURE_CONSISTENCY
    removeMissingFileTracks( ipod );
#endif

    checkForUnreferencedFiles( ipod, stream );

    stream << "</body></html>\n";
}

