/***************************************************************************
 *   Copyright (C) 2004 by Michael Schulze                                 *
 *   mike.s@genion.de                                                      *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/
#include "kpod.h"

#include <kdebug.h>

using namespace itunesdb;


KPod::KPod( const QString& ipodBase )
    : IPod( ipodBase )
{
    createdArtists.setAutoDelete( true );
}


KPod::~KPod()
{
}

bool KPod::open() {
    kdDebug() << "IPod::open(" << getBasePath() << ");" << endl;

    if ( !IPod::open() ) {
        return false;
    }

    logfileentrypos = 0;
    replayLog();

    return true;
}

void KPod::close() {
    flushLog();
    IPod::close();
}

void KPod::synchronize() {
    IPod::synchronize();
    flushLog();
}

bool KPod::ensureConsistency() {
    kdDebug() << "KPod::ensureConsistency()" << endl;
    if( !isStillConnected() ) {
        flushLog();
        return false;
    }

    replayLog();

    return true;
}


KPod::IPodError KPod::deleteArtist(const QString& artistname, bool log) {
    QStringList * albumlist = createdArtists.find( artistname );
    if ( albumlist ) {
        if ( !albumlist->isEmpty() ) {
            albumlist->clear();
        }
        createdArtists.remove( artistname );
    }

    // don't care if this fails since the only error is Err_DoesNotExist
    getITunesDB().removeArtist( artistname );

    if (log) {
        QStringList actions;
        actions << artistname;
        appendLogEntry( ACT_DELETE_ARTIST, actions );
    }

    return Err_None;
}

/*!
\fn KPod::renameAlbum(const QString& name, const QString& newname)
 */ 
KPod::IPodError KPod::renameAlbum(const QString& artistname, const QString& title, const QString& newartistname, const QString& newtitle, bool log)
{
    kdDebug() << "KPod::renameAlbum() " << title << endl;
    
    ITunesDB & itunesdb = getITunesDB();

    if ( !itunesdb.isOpen() ) {
        return Err_NotOpen;
    }
    if ( itunesdb.containsAlbum( newartistname, newtitle ) ) {    // does it already exist?
        return Err_AlreadyExists;
    }

    if( !itunesdb.renameAlbum( artistname, title, newartistname, newtitle ) ) {
        kdDebug() << "KPod::renameAlbum() issued an internal error" << endl;
        return Err_Internal;
    }
    
    // TODO rename album in createdArtists
    
    // check, if no more tracks with this artis exist and create an entry in createdArtists
    if ( !itunesdb.containsArtist( artistname ) ) {
        createArtist( artistname, false );
		
    }

    if (log) {
        QStringList actions;
        actions << artistname << title << newartistname << newtitle;
        appendLogEntry(ACT_RENAME_ALBUM, actions);
    }

    kdDebug() << "KPod::renameAlbum() finished" << endl;
    return Err_None;
}


KPod::IPodError KPod::renameArtist(const QString& artistname, const QString& newartistname, bool log) {
    kdDebug() << "KPod::renameArtist() " << artistname << endl;
    if (!getITunesDB().isOpen()) {
        return Err_NotOpen;
    }

    if ( artistname == newartistname ) {
        return Err_None;
    }

    ITunesDB::ArtistIterator trackIter = getITunesDB().getTracksByArtist( artistname );
    if ( !trackIter.hasNext() ) {
        QStringList * albums = createdArtists.take( artistname );
        if ( albums ) {
            createdArtists.insert( newartistname, albums );
        } else {
            // ignore
        }
    }

    while ( trackIter.hasNext() ) {
        trackIter.next()->setArtist( newartistname );
    }

    if ( log ) {
        QStringList actions;
        actions << artistname << newartistname;
        appendLogEntry( ACT_RENAME_ARTIST, actions );
    }

    kdDebug() << "KPod::renameArtist() finished" << endl;
    return Err_None;
}


KPod::IPodError KPod::deleteAlbum(const QString& artistname, const QString& title, bool log) {
    // remove artist from empty artist name entries
    QStringList * albumlist = createdArtists.find( artistname );
    if ( albumlist && albumlist->contains( title ) ) {
        albumlist->remove( title );
    }

    // Get the album to delete
    // TODO use Iterator here
    TrackPtrList album;
    if ( !getAlbum( artistname, title, album ) ) {
        return Err_DoesNotExist;
    }

    // delete all tracks of the found album
    TrackPtrList::Iterator trackiter = album.iterator();
    while ( trackiter.hasNext() ) {
        itunesdb::Track * track = trackiter.next();
        if (track == NULL) {
            continue;
        }

        QString filename = getITunesDB().getFileForPathInfo( track->getPathInfo() );
        if ( QFile::exists( filename ) ) {
            QFile::remove( filename );
        }
        getITunesDB().removeTrack( track->getID() );
    }

    // check, if no more tracks with this artis exist and create an entry in createdArtists
    if ( !getITunesDB().containsArtist( artistname ) ) {
        createArtist( artistname, false );
    }

    if ( log ) {
        QStringList actions;
        actions << artistname << title;
        appendLogEntry(ACT_REM_ALBUM, actions);
        getSysInfo().refreshDiskUsageStats();
    }

    return Err_None;
}

TrackPtrList *
KPod::getTracksByArtist( const QString& artistname, TrackPtrList& buffer ) const {
    return getITunesDB().getTracksByArtist( artistname, buffer );
}

TrackPtrList *
KPod::getAlbum( const QString &artistname, const QString &albumname, TrackPtrList& buffer ) const {
	TrackPtrList * result = getITunesDB().getAlbum( artistname, albumname, buffer );
	if ( !result && containsAlbum( artistname, albumname ) ) {
		// find it in the createdArtists ...
		return &buffer;
	}
	return result;
}

/**
 * returns the Track found by the given information or NULL if no such Track could be found
 * @param artistname the name of the artist
 * @param albumname the name of the album
 * @param title the title of the track
 */
ITunesDBTrack * KPod::findTrack(const QString& artistname, const QString& albumname, const QString& title) const {
    return getITunesDB().findTrack( artistname, albumname, title );
}


ITunesDBTrack * KPod::findDuplicateOf( const Track& track ) const {
    // use findFirstTrackBy( TrackPredicates::ByFullInfo or derived ) if more options are needed
    return getITunesDB().findTrack(
            track.getArtist(), track.getAlbum(), track.getTitle(), track.getTrackNumber() );
}


/*!
\fn KPod::renamePlaylist(const QString& name, const QString& newname)
 */
KPod::IPodError KPod::renamePlaylist(const QString& title, const QString& newtitle, bool log)
{
    if( !getITunesDB().isOpen() ) {
        return Err_NotOpen;
    }
    if( getITunesDB().getPlaylistByTitle( newtitle ) ) {    // does it already exist?
        return Err_AlreadyExists;
    }
    
    itunesdb::Playlist * playlist = getITunesDB().getPlaylistByTitle( title );
    if( playlist == NULL) {
        return Err_DoesNotExist;
    }
    
    playlist->setTitle(newtitle);
    
    if (log) {
        QStringList actions;
        actions << title << newtitle;
        appendLogEntry( ACT_RENAME_PLAYLIST, actions);
    }
    
    return Err_None;
}


/*!
\fn KPod::deletePlaylist(const QString& name)
 */
KPod::IPodError KPod::deletePlaylist(const QString& title, bool log)
{
    if( !getITunesDB().removePlaylist( title, true ) ) {
        return Err_DoesNotExist;
    }
    if ( log ) appendLogEntry(ACT_REM_PLAYLIST, QStringList(title));

    return Err_None;
}


QStringList * KPod::getPlaylistTitles(QStringList& buffer) {
    return getITunesDB().getPlaylistTitles( buffer );
}


QStringList * KPod::getArtists( QStringList &buffer ) const {
    QStringList * result = getITunesDB().getArtists(buffer);
    QDict<QStringList> * tempArtists = const_cast<QDict<QStringList>*> (&createdArtists);
    if ( result ) {
        QDictIterator<QStringList> artistIter( *tempArtists );
        while ( artistIter ) {
            QString currentArtist = artistIter.currentKey();
            if ( !result->contains( currentArtist ) ) {
                result->append( currentArtist );
                ++artistIter;
            } else {
                if ( (*artistIter)->isEmpty() ) {
                    tempArtists->remove( currentArtist );
                } else {
                    ++artistIter;
                }
            }
        }
    }
    return result;
}


QStringList * KPod::getAlbumsByArtist( const QString& artistName, QStringList& buffer ) const 
{
    QStringList * itunesDBResult = getITunesDB().getAlbumsByArtist( artistName, buffer );

    QDict<QStringList> * tempArtists = const_cast<QDict<QStringList>*> (&createdArtists);
    QStringList * albums = tempArtists->find( artistName );
    if ( albums == NULL ) {
        return itunesDBResult;
    }
    
    QStringList::iterator albumIter = albums->begin();
    while ( albumIter != albums->end() ) {
        if ( !buffer.contains( *albumIter ) ) {
            buffer.append( *albumIter );
            ++albumIter;
        } else {
            albumIter = albums->remove( albumIter );
        }
    }
    return &buffer;
}


bool KPod::containsArtist( const QString& artistname ) const {
    QStringList * albumList = createdArtists.find( artistname );
    if ( albumList ) {
        // just created - no tracks yet
        return true;
    }
    return getITunesDB().containsArtist( artistname );
}

bool KPod::containsAlbum( const QString& artistname, const QString& albumname ) const {
    QStringList * albumlist = createdArtists.find( artistname );
    if ( albumlist && albumlist->contains( albumname ) ) {
        // just created - no tracks yet
        return true;
    }
    return getITunesDB().containsAlbum( artistname, albumname );
}


KPod::IPodError KPod::createPlaylist(const QString& playlisttitle, bool log) {
    if( getITunesDB().getPlaylistByTitle( playlisttitle ) != NULL ) {
        return Err_AlreadyExists;
    }

    ITunesDBPlaylist * playlist = getITunesDB().createNewPlaylist( playlisttitle );
    getITunesDB().addPlaylist( playlist );

    if ( log ) {
        appendLogEntry( ACT_ADD_PLAYLIST, QStringList( playlist->getTitle() ) );
    }

    return Err_None;
}

KPod::IPodError KPod::createArtist( const QString& artistname, bool log ) {
    if ( getITunesDB().containsArtist( artistname ) || createdArtists.find( artistname ) ) {
        return Err_AlreadyExists;
    }

    createdArtists.insert( artistname, new QStringList() );

    if ( log ) {
        appendLogEntry( ACT_CREATE_ARTIST, QStringList( artistname ));
    }

    return Err_None;
}

KPod::IPodError KPod::createAlbum(const QString& artistName, const QString& albumName, bool log ) {
    QStringList * albums = createdArtists.find( artistName );
    if ( !albums ) {
        albums = new QStringList();
        createdArtists.insert( artistName, albums );
    }

    if ( albums->contains( albumName ) ) {
        return Err_AlreadyExists;
    }
    albums->append( albumName );

    if ( log ) {
        QStringList actions;
        actions << artistName << albumName;
        appendLogEntry( ACT_CREATE_ALBUM, actions);
    }

    return Err_None;
}


KPod::IPodError KPod::addTrackToPlaylist(const itunesdb::Track& track, const QString& playlisttitle, bool log) {
    itunesdb::Playlist * playlist = getITunesDB().getPlaylistByTitle(playlisttitle);
    if(playlist == NULL) {
        return Err_DoesNotExist;
    }

    playlist->addPlaylistItem(track.getID());

    if (log) {
        QStringList actions;
        actions << playlist->getTitle() << QString::number( track.getID(), 36 );
        appendLogEntry(ACT_ADD_TO_PLAYLIST, actions);
    }

    return Err_None;
}

KPod::IPodError KPod::removeFromPlaylist(Q_UINT32 position, const QString& playlisttitle, bool log) {
    itunesdb::Playlist * playlist = getITunesDB().getPlaylistByTitle(playlisttitle);
    if ( playlist == NULL ) {
        return Err_DoesNotExist;
    }

    playlist->removeTrackAt( position );

    if ( log ) {
        QStringList actions;
        actions << playlist->getTitle() << QString::number( position, 36 );
        appendLogEntry( ACT_REM_FROM_PLAYLIST, actions );
    }

    return Err_None;
}


ITunesDBTrack * KPod::getTrackByID(const Q_UINT32 id) const {
    return getITunesDB().getTrackByID(id);
}


KPod::IPodError KPod::addTrack( ITunesDBTrack * track, bool log ) {
    bool result = getITunesDB().addTrack( track );

    if ( result && log ) {
        QStringList actions;
        actions = trackToLogEntry( *track, actions );
        appendLogEntry( ACT_ADD_TRACK, actions );
        getSysInfo().refreshDiskUsageStats();
    }

    return result ? Err_None : Err_Operation_Failed;
}

KPod::IPodError KPod::moveTrack(itunesdb::Track& track, const QString& newartist, const QString& newalbum, bool log) {
    track.setArtist( newartist );
    track.setAlbum( newalbum );

    if (log) {
        QStringList actions;
        actions << QString::number( track.getID(), 36 ) << newartist << newalbum;
        appendLogEntry(ACT_MOV_TRACK, actions);
    }

    return Err_None;
}

KPod::IPodError KPod::deleteTrack(Q_UINT32 trackid, bool log) {
    if (!getITunesDB().removeTrack(trackid, true)) {
        return Err_DoesNotExist;
    }

    if (log) {
        QStringList actions;
        actions << QString::number( trackid, 36 );
        appendLogEntry(ACT_REM_TRACK, actions);
        getSysInfo().refreshDiskUsageStats();
    }

    return Err_None;
}

/*!
\fn KPod::trackToLogEntry()
 */
QStringList& KPod::trackToLogEntry( const itunesdb::Track& track, QStringList& valuebuffer ) const {
    valuebuffer << QString().setNum( track.getID(), 36 );
    valuebuffer << track.getPathInfo();
    valuebuffer << track.getArtist();
    valuebuffer << track.getAlbum();
    valuebuffer << track.getTitle();
    valuebuffer << track.getComment();
    valuebuffer << track.getGenre();
    valuebuffer << track.getComposer();
    valuebuffer << QString().setNum( track.getYear(), 36 );
    valuebuffer << QString().setNum( track.getFileSize(), 36 );
    valuebuffer << QString().setNum( track.getLastModified(), 36 );
    valuebuffer << QString().setNum( track.getBitrate(), 36 );
    valuebuffer << QString().setNum( track.getSamplerate(), 36 );
    valuebuffer << QString().setNum( track.getTrackLength(), 36 );
    valuebuffer << QString().setNum( track.getTrackNumber(), 36 );
    valuebuffer << QString().setNum( track.getCdNumber(), 36 );
    valuebuffer << QString().setNum( track.getCdCount(), 36 );
    valuebuffer << QString().setNum( track.getVolumeAdjust(), 36 );
    valuebuffer << QString().setNum( track.getVbrFlag(), 36 );
    valuebuffer << QString().setNum( track.getMediaType(), 36 );
    valuebuffer << QString().setNum( track.getCompilationFlag(), 36 );
    valuebuffer << QString().setNum( track.getRating(), 36 );
    valuebuffer << QString().setNum( track.getPlayCount(), 36 );
    valuebuffer << QString().setNum( track.getDBID(), 36 );
    valuebuffer << track.getCategory();
    valuebuffer << track.getDescription();
    valuebuffer << track.getEqualizerSetting();
    valuebuffer << track.getFileFormatDesc();
    valuebuffer << QString().setNum( track.getBPM(), 36 );
    valuebuffer << track.getGrouping();
    valuebuffer << QString().setNum( track.getStartTime(), 36 );
    valuebuffer << QString().setNum( track.getStopTime(), 36 );
    valuebuffer << QString().setNum( track.getSoundCheck(), 36 );
    valuebuffer << track.getAlbumArtist();
    valuebuffer << track.getKeywords();
    valuebuffer << QString().setNum( track.getLastSkippedTime(), 36 );
    valuebuffer << QString().setNum( track.hasLyrics() ? 1 : 0, 36 );
    valuebuffer << QString().setNum( track.getNumPregapSamples(), 36 );
    valuebuffer << QString().setNum( track.getNumSongSamples(), 36 );
    valuebuffer << QString().setNum( track.getNumPostgapSamples(), 36 );
    valuebuffer << QString().setNum( track.getGaplessDataSize(), 36 );
    valuebuffer << track.getAlbumForSort();
    valuebuffer << track.getArtistNameForSort();
    valuebuffer << track.getAlbumArtistForSort();
    valuebuffer << track.getTitleForSort();
    valuebuffer << track.getTvShow();
    valuebuffer << QString().setNum( track.getSeasonNumber(), 36 );
    valuebuffer << QString().setNum( track.getEpisodeNumber(), 36 );

    return valuebuffer;
}


/*!
\fn KPod::appendLogEntry( int type, QStringList& values)
 */
bool KPod::appendLogEntry(KPod::LogActionType type, const QStringList& values)
{
    bool _unlock_ = false;
    QFile logfile(getLogfileName());
    if (!logfile.open(IO_ReadWrite | IO_Append)) {
        return false;
    }
    
    if (!isLocked()) {
        lock(true);
        _unlock_ = true;
    }
    
    QByteArray logentry;
    QDataStream stream(logentry, IO_WriteOnly);
    stream.setByteOrder(QDataStream::LittleEndian);
    for (QStringList::const_iterator value_it = values.constBegin(); value_it != values.constEnd(); ++value_it) {
        stream << *value_it;
    }
    
    QDataStream logstream(&logfile);
    logstream.setByteOrder(QDataStream::LittleEndian);
    logstream << type;
    logstream << logentry;
    
    logstream.unsetDevice();
    
    logfileentrypos++;
    
    logfile.flush();
    logfile.close();
    
    if(_unlock_) {
        unlock();
    }

    return true;
}



/*!
\fn TrackMetadata::readFromLogEntry()
 */
class TrackLogEntryReader {
    const QStringList & values;
public:
TrackLogEntryReader( const QStringList & fields ) : values( fields ) {}
    virtual ~TrackLogEntryReader() {}
    
    const QString& getArtist() { return values[ 2 ]; }
    const QString& getAlbum() { return values[ 3 ]; }
    const QString& getTitle() { return values[ 4 ]; }
    const QString& getComment() { return values[ 5 ]; }
    const QString& getGenre() { return values[ 6 ]; }
    const QString& getComposer() { return values[ 7 ]; }
    Q_UINT32 getYear() { return values[ 8 ].toUInt( NULL, 36 ); }
    Q_UINT32 getFileSize() { return values[ 9 ].toUInt( NULL, 36 ); }
    Q_UINT32 getLastModified() { return values[ 10 ].toUInt( NULL, 36 ); }
    Q_UINT32 getBitrate() { return values[ 11 ].toUInt( NULL, 36 ); }
    Q_UINT32 getSamplerate() { return values[ 12 ].toUInt( NULL, 36 ); }
    Q_UINT32 getTrackLength() { return values[ 13 ].toUInt( NULL, 36 ); }
    Q_UINT32 getTrackNumber() { return values[ 14 ].toUInt( NULL, 36 ); }
    Q_UINT32 getCdNumber() { return values[ 15 ].toUInt( NULL, 36 ); }
    Q_UINT32 getCdCount() { return values[ 16 ].toUInt( NULL, 36 ); }
    Q_UINT32 getVolumeAdjust() { return values[ 17 ].toUInt( NULL, 36 ); }
    unsigned char getVbrFlag() { return values[ 18 ].toUShort( NULL, 36 ) & 255; }
    itunesdb::Track::MediaType getMediaType() { return (itunesdb::Track::MediaType)((char)values[ 19 ].toUInt( NULL, 36 ) & 255); }
    unsigned char getCompilationFlag() { return values[ 20 ].toUShort( NULL, 36 ) & 255; }
    unsigned char getRating() { return values[ 21 ].toUShort( NULL, 36 ) & 255; }
    Q_UINT32 getPlayCount() { return values[ 22 ].toUInt( NULL, 36 ); }
    Q_UINT32 getLastPlayed() { return 0; }
    Q_UINT32 getFileFormatCode() { return 0xC; }
    Q_UINT32 getTrackCount() { return 0; }
    const QString& getCategory() { return values[ 24 ]; }
    const QString& getDescription() { return values[ 25 ]; }
    const QString& getEqualizerSetting()  { return values[ 26 ]; }
    const QString& getFileFormatDesc()  { return values[ 27 ]; }
    const Q_UINT16 getBPM() { return values[ 28 ].toUInt( NULL, 36 ); }
    const QString& getGrouping() { return values[ 29 ]; }
    const Q_UINT32 getStartTime() { return values[ 30 ].toUInt( NULL, 36 ); }
    const Q_UINT32 getStopTime() { return values[ 31 ].toUInt( NULL, 36 ); }
    const Q_UINT32 getSoundCheck() { return values[ 32 ].toUInt( NULL, 36 ); }
    const QString& getAlbumArtist() { return values[ 33 ]; }
    const QString& getKeywords() { return values[ 34 ]; }
    const Q_UINT32 getLastSkippedTime() { return values[ 35 ].toUInt( NULL, 36 ); }
    const Q_UINT32 getSkipCount() { return values[ 36 ].toUInt( NULL, 36 ); }
    const bool hasLyrics() { return values[ 37 ].toUShort( NULL, 36 ); }
    const Q_UINT32 getNumPregapSamples() { return values[ 38 ].toUInt( NULL, 36 ); }
    const Q_UINT64 getNumSongSamples() { return values[ 39 ].toULong( NULL, 36 ); }
    const Q_UINT32 getNumPostgapSamples() { return values[ 40 ].toUInt( NULL, 36 ); }
    const Q_UINT32 getGaplessDataSize() { return values[ 41 ].toUInt( NULL, 36 ); }
    const QString& getAlbumForSort() { return values[ 42 ]; }
    const QString& getArtistNameForSort() { return values[ 43 ]; }
    const QString& getAlbumArtistForSort() { return values[ 44 ]; }
    const QString& getTitleForSort() { return values[ 45 ]; }
    const QString& getTvShow() { return values[ 46 ]; }
    const Q_UINT32 getSeasonNumber() { return values[ 47 ].toUInt( NULL, 36 ); }
    const Q_UINT32 getEpisodeNumber() { return values[ 48 ].toUInt( NULL, 36 ); }
};
ITunesDBTrack * KPod::readTrackFromLogEntry( const QStringList& values ) {
    bool conversionResult = true;
    // some sanity checks first
    if ( values.size() < 28 ) {
        return false;    // wtf?
    }

    Q_UINT32 trackID = values[ 0 ].toUInt( &conversionResult, 36 );
    if ( !conversionResult || trackID == 0 ) {
        return NULL;
    }

    ITunesDBTrack * result = getITunesDB().createEmptyTrack( trackID );
    result->setPathInfo( values[ 1 ] );
    result->setDBID( values[ 23 ].toULong( &conversionResult, 36 ) );

    TrackLogEntryReader logEntryReader( values );
    result->readFrom( logEntryReader );

    return result;
}

/*!
\fn kio_ipodslaveProtocol::replayLog()
 */
void KPod::replayLog()
{
    kdDebug() << "KPod::replayLog()" << endl;
    bool _unlock_ = false;
    if (!isLocked()) {
        lock(false);
        _unlock_ = true;
    }
    
    // kdDebug() << "KPod::replayLog() locked!" << endl;
    
    QFile logfile(getLogfileName());
    if (!logfile.open(IO_ReadOnly)) {
        if(_unlock_)
            unlock();
        return;
    }
    
    replayingLog = true;
    // kdDebug() << "KPod::replayLog() logfile opened!" << endl;
    
    QDataStream logstream(&logfile);
    logstream.setByteOrder(QDataStream::LittleEndian);
    
    // ignore the changes we already know about
    for (uint i= 0; i< logfileentrypos; i++) {
        Q_UINT32 type;
        QByteArray buffer;
        if( logstream.atEnd()) {    // ick
            logfileentrypos= i;
            break;
        }
        logstream >> type;
        logstream >> buffer;
    }
    
    while (!logstream.atEnd()) {    // read new log entries
        QByteArray entrydata;
        Q_UINT32 type;
        QStringList values;
        
        logstream >> type;
        logstream >> entrydata;
        
        if (type >= NUM_ACTIONS) {
            continue;
        }
        
        logfileentrypos++;
        
        if( entrydata.isEmpty())
            continue;
        
        // parse entry elements
        QDataStream entrystream(entrydata, IO_ReadOnly);
        entrystream.setByteOrder( QDataStream::LittleEndian);
        while(!entrystream.atEnd()) {
            QString value;
            entrystream >> value;
            values.push_back(value);
        }
        
        switch( type) {    // handle logfile entry
        case ACT_ADD_PLAYLIST:
            // add playlist
            if (values.size()> 0) {
                createPlaylist(values[0], false);
            }
            break;
        case ACT_REM_PLAYLIST:
            if (values.size()> 0) {
                deletePlaylist( values[ 0], false);
            }
            break;
        case ACT_RENAME_PLAYLIST:
            if (values.size()> 1) {
                renamePlaylist(values[0], values[1], false);
            }
            break;
        case ACT_REM_ALBUM:
            if (values.size()> 1) {
                deleteAlbum(values[0], values[1], false);
            }
            break;
        case ACT_RENAME_ALBUM:
            if (values.size()> 3) {
                renameAlbum(values[0], values[1], values[2], values[3], false);
            }
            break;
        case ACT_ADD_TO_PLAYLIST: {
            if ( values.size() > 1 ) {
                bool conversion_ok= true;
                Q_UINT32 trackid= values[ 1].toUInt( &conversion_ok, 36 );
                if(conversion_ok) {
                    itunesdb::Track * track = getTrackByID( trackid);
                    if(track != NULL)
                        addTrackToPlaylist(*track, values[0], false);
                }
            }
        }
            break;
        case ACT_REM_FROM_PLAYLIST:
            if (values.size()> 1) {
                bool conversion_ok = true;
                int tracknum = values[ 1].toUInt( &conversion_ok, 36 );
                if(conversion_ok)
                    removeFromPlaylist(tracknum, values[0], false);
            }
            break;
        case ACT_ADD_TRACK:
            if (values.size() > 0) {
                ITunesDBTrack * track = readTrackFromLogEntry( values );
                if ( track ) {
                    addTrack( track, false );
                }
            }    // ignore otherwise
            break;
        case ACT_MOV_TRACK:
            if (values.size() > 2) {
                bool conversion_ok = true;
                int trackid = values[0].toUInt( &conversion_ok, 36 );
                if (!conversion_ok)
                    break;
                itunesdb::Track * track = getTrackByID(trackid);
                if (track == NULL)
                    break;
                moveTrack(*track, values[1], values[2], false);
            }
            break;
        case ACT_REM_TRACK: {
            if (values.size() > 0) {
                bool conversion_ok = true;
                int trackid = values[ 0].toUInt( &conversion_ok, 36 );
                if(conversion_ok)
                    deleteTrack(trackid, false);
            }    // ignore otherwise
        }
            break;
        case ACT_DELETE_ARTIST: {
            if (values.size() > 0) {
                deleteArtist(values[0], false);
            }    // ignore otherwise
        }
            break;
        case ACT_RENAME_ARTIST: {
            if ( values.size() > 1 ) {
                renameArtist( values[ 0 ], values[ 1 ], false );
            }
        }
            break;
        case ACT_CREATE_ARTIST: {
            if ( values.size() > 0 ) {
                createArtist( values[ 0 ], false );
            }
        }
            break;
        case ACT_CREATE_ALBUM: {
            if ( values.size() > 1 ) {
                createAlbum( values[ 0 ], values[ 1 ], false );
            }
        }
            break;
        default:
            break;
        }
    }
    
    replayingLog = false;
    
    if(_unlock_) {
        unlock();
    }
}


/*!
\fn kio_ipodslaveProtocol::flushLog
 */
void KPod::flushLog()
{
    if( QFile::exists(getLogfileName())) {
        QFile::remove(getLogfileName());
    }
    logfileentrypos = 0;
}

uint KPod::getNumPlaylists() { return getITunesDB().getNumPlaylists(); }

uint KPod::getNumTracks() { return getITunesDB().getNumTracks(); }

QString KPod::getITunesDbFilename() { return getITunesDB().getFilename(); }

QString KPod::getLogfileName() const { return getBasePath() + LOGFILEPREFIX + QString::number((unsigned int) ( getITunesDB().lastModified() / 1000 ) ); }
