/***************************************************************************
 *   Copyright (C) 2004 by Michael Schulze                                 *
 *   mike.s@genion.de                                                      *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/
#ifndef KPOD_H
#define KPOD_H

#include <libqtpod/ipod.h>
#include <qdict.h>

#define LOG_DEFAULT true

/**
KPods representation of an IPod


@author Michael Schulze
*/
class KPod : public IPod
{
public:
    /**
     * Error values returned by certain functions.
     */
    enum IPodError {
        Err_None = 0,
        Err_ParseError,
        Err_NotOpen,
        Err_AlreadyExists,
        Err_DoesNotExist,
        Err_NotEmpty,
        Err_Internal,
        Err_Operation_Failed
    };

    enum LogActionType {
        ACT_ADD_PLAYLIST= 0,
            ACT_REM_PLAYLIST,
            ACT_RENAME_PLAYLIST,
            ACT_REM_ALBUM,
            ACT_RENAME_ALBUM,
            ACT_ADD_TO_PLAYLIST,
            ACT_REM_FROM_PLAYLIST,
            ACT_ADD_TRACK,
            ACT_MOV_TRACK,
            ACT_REM_TRACK,
            ACT_DELETE_ARTIST,
            ACT_RENAME_ARTIST,
            ACT_CREATE_ARTIST,
            ACT_CREATE_ALBUM,
            NUM_ACTIONS
    };

    KPod( const QString& ipodBase );
    virtual ~KPod();

    /**
     * Tries to open an ipod. That means reading all important information from the device.
     * @return true if successful, otherwise false
     */
    virtual bool open();

    /**
     * Closes an open iPod
     */
    virtual void close();

    /**
     * Writes the ITunesDB and other changed data to the ipods disk
     */
    virtual void synchronize();

    /**
     * Makes sure that data is consistent between instances opened on the same IPod and the IPod is still available
     */
    bool ensureConsistency();

    /**
     * removes an empty artist
     */
    IPodError deleteArtist(const QString& artistname, bool log = LOG_DEFAULT);

    /**
     * Renames an album with the given title and artist to newtitle, newartist
     * @return AlreadyExists if an album with the given newtitle already exists or DoesNotExist if an album with the given title doesn't exist
     */
    IPodError renameAlbum(const QString& artistname, const QString& title, const QString& newartistname, const QString& newtitle, bool log = LOG_DEFAULT);
    
    /**
     * Renames an artist
     */
    IPodError renameArtist(const QString& artistname, const QString& newartistname, bool log = LOG_DEFAULT);
    
    /**
     * removes an album and all it's contents from the database
     */
    IPodError deleteAlbum(const QString& artistname, const QString& title, bool log = LOG_DEFAULT);
    
    /**
     * Creates a new playlist with the given title
     * @return Err_AlreadyExists if a playlist with the given title already exists
     */
    IPodError createPlaylist(const QString& playlisttitle, bool log = LOG_DEFAULT);
    
    /**
     * Creates a new Artist entry
     * This function is not really an ipod usecase but an belongs to the slave side
     */
    IPodError createArtist( const QString& artistname, bool log = LOG_DEFAULT );
    
    /**
     * Creates a new Album entry
     * This function is not really an ipod usecase but an belongs to the slave side
     */
    IPodError createAlbum( const QString& artist, const QString& albumName, bool log = LOG_DEFAULT );
    
    /**
     * Renames the playlist with the given title to newtitle.
     * @return AlreadyExists if a playlist with the given newtitle already exists or DoesNotExist if a playlist with the given title doesn't exist
     */
    IPodError renamePlaylist(const QString& title, const QString& newtitle, bool log = LOG_DEFAULT);
    
    /**
     * deletes the playlist with the given name from the database and frees the memory resources
     * ATTENTION: pointers to the playlist with the given name will become invalid
     * @return DoesNotExist if a playlisz with the given title does not exist
     */
    IPodError deletePlaylist(const QString& title, bool log = LOG_DEFAULT);
    
    /**
     * Removes the Track with the given trackid from the database.
     * This also removes all playlist references to this track
     * @param trackid ID of the track to be deleted
     * @return the trackid of the Track or 0 if no such Track exists
     */
    IPodError deleteTrack(Q_UINT32 trackid, bool log = LOG_DEFAULT);
    
    /**
     * moves the given track from one album to another.
     * This method may be not accessible in the future when/if track::setAlbum() track::setArtist() work correctly
     */
    IPodError moveTrack(itunesdb::Track& track, const QString& newartist, const QString& newalbum, bool log = LOG_DEFAULT);

    /**
     * Fills the given QStringList with the names of all playlists in the database
     * @return a pointer to the given StringList
     */
    QStringList * getPlaylistTitles(QStringList &buffer);

    /**
     * Fills the given QStringList with the names of all artists in the database
     * @return a pointer to the StringList
     */
    QStringList * getArtists( QStringList &buffer) const;

    /**
     * Fills the given QStringList with the names of the albums by the given artistmap
     * @return a pointer to the QStringList or NULL if no such artist exists
     */
    QStringList * getAlbumsByArtist( const QString& artistname, QStringList& buffer ) const;

    /**
     * Returns true if the given artist exists
     */
    bool containsArtist( const QString& artistname ) const;

    /**
     * Returns true if the given album exists
     */
    bool containsAlbum( const QString& artistname, const QString& albumname ) const;

    /**
     * adds a new Track to the database after itunesDBlistener is done (finished)
     */
    IPodError addTrack(ITunesDBTrack * track, bool log = LOG_DEFAULT);

    /**
     * Adds the track with the given trackid to the playlist with the given title.
     */
    IPodError addTrackToPlaylist(const itunesdb::Track& track, const QString& playlisttitle, bool log = LOG_DEFAULT);

    /**
     * Removes the track at the given position from the given playlist
     */
    IPodError removeFromPlaylist(Q_UINT32 position, const QString& playlisttitle, bool log = LOG_DEFAULT);

    /**
     * returns the Track corresponding to the given ID
     * @param id ID of the track
     * @return the Track corresponding to the given ID
     */
    ITunesDBTrack * getTrackByID( const Q_UINT32 id) const;

    /**
     * Returns all Tracks by the given artist
     * FIXME create delegate method returning an iterator
     */
    itunesdb::TrackPtrList *
        getTracksByArtist( const QString& artistname, itunesdb::TrackPtrList& buffer ) const;

    /**
     * Returns an album (list of trackIDs) by the given Artist/Album
     * @param artistname name of the artist of the album in question
     * @param albumname name of the album
     * @return a pointer to the album or NULL if no album could by found
     * FIXME create delegate method returning an iterator
     */
    itunesdb::TrackPtrList * getAlbum(const QString &artistname, const QString &albumname, itunesdb::TrackPtrList& buffer) const;

    /**
     * returns the Track found by the given information or NULL if no such Track could be found
     * @param artistname the name of the artist
     * @param albumname the name of the album
     * @param title the title of the track
     */
    ITunesDBTrack * findTrack(const QString& artistname, const QString& albumname, const QString& title) const;

    /**
     * returns the duplicate of the given Track if there is one or NULL if no duplicate could be found. The current implementation looks for tracks with the same artist/album/title/tracknum.
     * @param track the track to find the duplicate of
     * @return the track found or NULL if no such track exists.
     */
    ITunesDBTrack * findDuplicateOf( const itunesdb::Track& track ) const;

    /**
     * Returns the number of playlists found on the iPod.
     */
    uint getNumPlaylists();

    /**
     * Returns the number of tracks found on the iPod.
     */
    uint getNumTracks();
    
    bool isDirty() const {
    	return IPod::isDirty() || QFile ( getLogfileName() ).exists();
    }

    /**
     * Returns the full path to the iPods database file.
     */
    QString getITunesDbFilename();

protected:
    QFile logfile;

    QString getLogfileName() const;

    bool appendLogEntry(KPod::LogActionType type, const QStringList& values);
    void replayLog();
    void flushLog();

private:
    QStringList& trackToLogEntry( const itunesdb::Track& track, QStringList& valuebuffer ) const;
    ITunesDBTrack * readTrackFromLogEntry( const QStringList& values );

    // TODO when removing all tracks of an artist we need to create an entry in here
    QDict<QStringList> createdArtists;

    bool replayingLog;
    uint logfileentrypos;

};

#endif
